# This file is part of the MapProxy project.
# Copyright (C) 2012, 2013 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import hashlib

try:
    import json
    json
except ImportError:
    json = None

try:
    import requests
    requests
except ImportError:
    requests = None

from mapproxy.client.log import log_request
from mapproxy.cache.tile import TileCreator, Tile
from mapproxy.source import SourceError
from mapproxy.util.lock import LockTimeout


def has_renderd_support():
    if not json or not requests:
        return False
    return True


class RenderdTileCreator(TileCreator):
    def __init__(self, renderd_address, tile_mgr, dimensions=None, priority=100, tile_locker=None):
        TileCreator.__init__(self, tile_mgr, dimensions)
        self.tile_locker = tile_locker.lock or self.tile_mgr.lock
        self.renderd_address = renderd_address
        self.priority = priority

    def _create_single_tile(self, tile):
        with self.tile_locker(tile):
            if not self.is_cached(tile):
                self._create_renderd_tile(tile.coord)
            self.cache.load_tile(tile)
        return [tile]

    def _create_meta_tile(self, meta_tile):
        main_tile = Tile(meta_tile.main_tile_coord)
        with self.tile_locker(main_tile):
            if not all(self.is_cached(t) for t in meta_tile.tiles if t is not None):
                self._create_renderd_tile(main_tile.coord)

        tiles = [Tile(coord) for coord in meta_tile.tiles]
        self.cache.load_tiles(tiles)
        return tiles

    def _create_renderd_tile(self, tile_coord):
        start_time = time.time()
        result = self._send_tile_request(self.tile_mgr.identifier, [tile_coord])
        duration = time.time()-start_time

        address = '%s:%s:%r' % (self.renderd_address,
                                self.tile_mgr.identifier, tile_coord)

        if result['status'] == 'error':
            log_request(address, 500, None, duration=duration, method='RENDERD')
            raise SourceError("Error from renderd: %s" % result.get('error_message', 'unknown error from renderd'))
        elif result['status'] == 'lock':
            log_request(address, 503, None, duration=duration, method='RENDERD')
            raise LockTimeout("Lock timeout from renderd: %s" % result.get(
                'error_message', 'unknown lock timeout error from renderd'))

        log_request(address, 200, None, duration=duration, method='RENDERD')

    def _send_tile_request(self, cache_identifier, tile_coords):
        identifier = hashlib.sha1(str((cache_identifier, tile_coords)).encode('ascii')).hexdigest()
        message = {
            'command': 'tile',
            'id': identifier,
            'tiles': tile_coords,
            'cache_identifier': cache_identifier,
            'priority': self.priority
        }
        try:
            resp = requests.post(self.renderd_address, data=json.dumps(message))
            return resp.json()
        except ValueError:
            raise SourceError("Error while communicating with renderd: invalid JSON")
        except requests.RequestException as ex:
            raise SourceError("Error while communicating with renderd: %s" % ex)



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mapproxy.client.http import HTTPClient
from mapproxy.client.wms import WMSInfoClient
from mapproxy.srs import SRS
from mapproxy.featureinfo import create_featureinfo_doc


class ArcGISClient(object):
    def __init__(self, request_template, http_client=None):
        self.request_template = request_template
        self.http_client = http_client

    def retrieve(self, query, format):
        url = self._query_url(query, format)
        resp = self.http_client.open(url)
        return resp

    def _query_url(self, query, format):
        req = self.request_template.copy()
        req.params.format = format
        req.params.bbox = query.bbox
        req.params.size = query.size
        req.params.bboxSR = query.srs
        req.params.imageSR = query.srs
        req.params.transparent = query.transparent

        return req.complete_url

    def combined_client(self, other, query):
        return


class ArcGISInfoClient(WMSInfoClient):
    def __init__(self, request_template, supported_srs=None, http_client=None,
                 return_geometries=False,
                 tolerance=5,
                 ):
        self.request_template = request_template
        self.http_client = http_client or HTTPClient()
        if not supported_srs and self.request_template.params.srs is not None:
            supported_srs = [SRS(self.request_template.params.srs)]
        self.supported_srs = supported_srs or []
        self.return_geometries = return_geometries
        self.tolerance = tolerance

    def get_info(self, query):
        if self.supported_srs and query.srs not in self.supported_srs:
            query = self._get_transformed_query(query)
        resp = self._retrieve(query)
        # always use query.info_format and not content-type from response (even esri example server aleays return
        # text/plain)
        return create_featureinfo_doc(resp.read(), query.info_format)

    def _query_url(self, query):
        req = self.request_template.copy()
        req.params.bbox = query.bbox
        req.params.size = query.size
        req.params.pos = query.pos
        req.params.srs = query.srs.srs_code
        if query.info_format.startswith('text/html'):
            req.params['f'] = 'html'
        else:
            req.params['f'] = 'json'

        req.params['tolerance'] = self.tolerance
        req.params['returnGeometry'] = str(self.return_geometries).lower()

        return req.complete_url



# This file is part of the MapProxy project.
# Copyright (C) 2025 Spatialys
#
# Initial development funded by Centre National d'Etudes Spatiales (CNES): https://cnes.fr
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import yaml

from mapproxy.config.loader import ProxyConfiguration
from mapproxy.script.gridconf_from_ogcapitilematrixset import (
    gridconf_from_ogcapitilematrixset_command,
)
from mapproxy.test.helper import capture
from mapproxy.test.http import mock_httpd


class TestUtilGridConfFromOGCAPITileMatrixSet(object):
    def test(self):
        landing_page = {
            "links": [
                {
                    "rel": "http://www.opengis.net/def/rel/ogc/1.0/tiling-schemes",
                    "type": "application/json",
                    "title": "Map tilesets available for this dataset (as JSON)",
                    "href": "/ogcapi/tileMatrixSets?f=json",
                },
            ]
        }

        tileMatrixSets = {
            "tileMatrixSets": [
                {
                    "id": "WebMercatorQuad",
                    "title": "WebMercatorQuad",
                    "links": [
                        {
                            "rel": "self",
                            "type": "application/json",
                            "title": "The JSON representation of the WebMercatorQuad tiling scheme definition",
                            "href": "/ogcapi/tileMatrixSets/WebMercatorQuad?f=json",
                        },
                    ],
                },
                {
                    "id": "invalid",
                    "title": "invalid",
                    "links": [
                        {
                            "rel": "self",
                            "type": "application/json",
                            "title": "The JSON representation of the invalid tiling scheme definition",
                            "href": "/ogcapi/tileMatrixSets/invalid?f=json",
                        },
                    ],
                },
            ]
        }

        WebMercatorQuad = {
            "id": "WebMercatorQuad",
            "title": "WebMercatorQuad",
            "uri": "http://www.opengis.net/def/tilematrixset/OGC/1.0/WebMercatorQuad",
            "crs": "http://www.opengis.net/def/crs/EPSG/0/3857",
            "orderedAxes": ["E", "N"],
            "wellKnownScaleSet": "http://www.opengis.net/def/wkss/OGC/1.0/GoogleMapsCompatible",
            "tileMatrices": [
                {
                    "id": "0",
                    "scaleDenominator": 559082264.028718,
                    "cellSize": 156543.03392804,
                    "cornerOfOrigin": "topLeft",
                    "pointOfOrigin": [-20037508.3427892, 20037508.3427892],
                    "matrixWidth": 1,
                    "matrixHeight": 1,
                    "tileWidth": 256,
                    "tileHeight": 256,
                },
                {
                    "id": "1",
                    "scaleDenominator": 279541132.01436,
                    "cellSize": 78271.5169640205,
                    "cornerOfOrigin": "topLeft",
                    "pointOfOrigin": [-20037508.3427892, 20037508.3427892],
                    "matrixWidth": 2,
                    "matrixHeight": 2,
                    "tileWidth": 256,
                    "tileHeight": 256,
                },
                {
                    "id": "2",
                    "scaleDenominator": 139770566.00718,
                    "cellSize": 39135.7584820102,
                    "cornerOfOrigin": "topLeft",
                    "pointOfOrigin": [-20037508.3427892, 20037508.3427892],
                    "matrixWidth": 4,
                    "matrixHeight": 4,
                    "tileWidth": 256,
                    "tileHeight": 256,
                },
            ],
        }

        invalid = copy.copy(WebMercatorQuad)
        invalid["id"] = "invalid"
        invalid["crs"] = "http://www.opengis.net/def/crs/EPSG/0/98765432"

        expected_reqs = [
            (
                {"path": r"/ogcapi"},
                {
                    "body": bytes(json.dumps(landing_page).encode("UTF-8")),
                    "headers": {"content-type": "application/json"},
                },
            ),
            (
                {"path": r"/ogcapi/tileMatrixSets?f=json"},
                {
                    "body": bytes(json.dumps(tileMatrixSets).encode("UTF-8")),
                    "headers": {"content-type": "application/json"},
                },
            ),
            (
                {"path": r"/ogcapi/tileMatrixSets/WebMercatorQuad?f=json"},
                {
                    "body": bytes(json.dumps(WebMercatorQuad).encode("UTF-8")),
                    "headers": {"content-type": "application/json"},
                },
            ),
            (
                {"path": r"/ogcapi/tileMatrixSets/invalid?f=json"},
                {
                    "body": bytes(json.dumps(invalid).encode("UTF-8")),
                    "headers": {"content-type": "application/json"},
                },
            ),
        ]
        with mock_httpd(("localhost", 42423), expected_reqs):
            with capture() as (out, err):
                args = ["dummy_utility", "http://localhost:42423/ogcapi"]
                gridconf_from_ogcapitilematrixset_command(args)
        captured_error = err.getvalue()
        captured_output = out.getvalue()
        assert (
            "Cannot handle http://localhost:42423/ogcapi/tileMatrixSets/invalid?f=json: "
            + "CRS http://www.opengis.net/def/crs/EPSG/0/98765432 is not supported"
            in captured_error
        )
        pos_grids = captured_output.find("grids:")
        config_dict = yaml.safe_load(captured_output[pos_grids:])
        assert config_dict == {
            "grids": {
                "WebMercatorQuad": {
                    "bbox": [
                        -20037508.3427892,
                        -20037508.342789043,
                        20037508.342789043,
                        20037508.3427892,
                    ],
                    "origin": "ul",
                    "res": [156543.03392804, 78271.5169640205, 39135.7584820102],
                    "srs": "EPSG:3857",
                    "tile_size": [256, 256],
                }
            }
        }

        config = ProxyConfiguration(config_dict)
        grid = config.grids["WebMercatorQuad"].tile_grid()
        assert grid.name == "WebMercatorQuad"
        assert grid.bbox == (
            -20037508.3427892,
            -20037508.342789043,
            20037508.342789043,
            20037508.3427892,
        )
        assert grid.origin == "ul"
        assert grid.srs.srs_code == "EPSG:3857"
        assert grid.tile_size == (256, 256)
        assert [grid.resolutions[level] for level in range(grid.levels)] == config_dict[
            "grids"
        ]["WebMercatorQuad"]["res"]



# This file is part of the MapProxy project.
# Copyright (C) 2011-2013 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import hashlib
import os
import threading
import time
from io import BytesIO
from itertools import groupby

from mapproxy.image import ImageSource
from mapproxy.cache.base import TileCacheBase, tile_buffer, REMOVE_ON_UNLOCK
from mapproxy.util.fs import ensure_directory
from mapproxy.util.lock import FileLock
from mapproxy.util.sqlite3 import sqlite3

import logging
log = logging.getLogger(__name__)

if not hasattr(glob, 'escape'):
    import re
    glob.escape = lambda pathname: re.sub(r'([*?[])', r'[\1]', pathname)


def sqlite_datetime_to_timestamp(datetime):
    if datetime is None:
        return None
    d = time.strptime(datetime, "%Y-%m-%d %H:%M:%S")
    return time.mktime(d)


class MBTilesCache(TileCacheBase):
    supports_timestamp = False

    def __init__(self, mbtile_file, with_timestamps=False, timeout=30, wal=False, ttl=0, coverage=None,
                 directory_permissions=None, file_permissions=None):
        super(MBTilesCache, self).__init__(coverage)
        md5 = hashlib.new('md5', mbtile_file.encode('utf-8'), usedforsecurity=False)
        self.lock_cache_id = 'mbtiles-' + md5.hexdigest()
        self.mbtile_file = mbtile_file
        self.directory_permissions = directory_permissions
        self.file_permissions = file_permissions
        self.supports_timestamp = with_timestamps
        self.ttl = with_timestamps and ttl or 0
        self.timeout = timeout
        self.wal = wal
        self.ensure_mbtile()
        self._db_conn_cache = threading.local()

    @property
    def db(self):
        if not getattr(self._db_conn_cache, 'db', None):
            self.ensure_mbtile()
            self._db_conn_cache.db = sqlite3.connect(self.mbtile_file, self.timeout)
        return self._db_conn_cache.db

    def cleanup(self):
        """
        Close all open connection and remove them from cache.
        """
        if getattr(self._db_conn_cache, 'db', None):
            self._db_conn_cache.db.close()
        self._db_conn_cache.db = None

    def ensure_mbtile(self):
        if not os.path.exists(self.mbtile_file):
            with FileLock(self.mbtile_file + '.init.lck', remove_on_unlock=REMOVE_ON_UNLOCK,
                          directory_permissions=self.directory_permissions, file_permissions=self.file_permissions):
                if not os.path.exists(self.mbtile_file):
                    ensure_directory(self.mbtile_file, self.directory_permissions)
                    self._initialize_mbtile()

    def _initialize_mbtile(self):
        log.info('initializing MBTile file %s', self.mbtile_file)
        with sqlite3.connect(self.mbtile_file) as db:
            if self.wal:
                db.execute('PRAGMA journal_mode=wal')

            stmt = """
                CREATE TABLE tiles (
                    zoom_level integer,
                    tile_column integer,
                    tile_row integer,
                    tile_data blob
            """

            if self.supports_timestamp:
                stmt += """
                    , last_modified datetime DEFAULT (datetime('now','localtime'))
                """
            stmt += """
                );
            """
            db.execute(stmt)

            db.execute("""
                CREATE TABLE metadata (name text, value text);
            """)
            db.execute("""
                CREATE UNIQUE INDEX idx_tile on tiles
                    (zoom_level, tile_column, tile_row);
            """)
            db.commit()

        if self.file_permissions:
            permission = int(self.file_permissions, base=8)
            log.info("setting file permissions on MBTile file: %s", permission)
            os.chmod(self.mbtile_file, permission)

    def update_metadata(self, name='', description='', version=1, overlay=True, format='png'):
        self.db.execute("""
            CREATE TABLE IF NOT EXISTS metadata (name text, value text);
        """)
        self.db.execute("""DELETE FROM metadata;""")

        if overlay:
            layer_type = 'overlay'
        else:
            layer_type = 'baselayer'

        self.db.executemany("""
            INSERT INTO metadata (name, value) VALUES (?,?)
            """,
                       (
                           ('name', name),
                           ('description', description),
                           ('version', version),
                           ('type', layer_type),
                           ('format', format),
                       )
                       )
        self.db.commit()

    def is_cached(self, tile, dimensions=None):
        if tile.coord is None:
            return True
        if tile.source:
            return True

        return self.load_tile(tile, dimensions=dimensions)

    def store_tile(self, tile, dimensions=None):
        if tile.stored:
            return True
        return self._store_bulk([tile])

    def store_tiles(self, tiles, dimensions=None):
        tiles = [t for t in tiles if not t.stored]
        return self._store_bulk(tiles)

    def _store_bulk(self, tiles):
        records = []
        # tile_buffer (as_buffer) will encode the tile to the target format
        # we collect all tiles before, to avoid having the db transaction
        # open during this slow encoding
        for tile in tiles:
            with tile_buffer(tile) as buf:
                content = buf.read()
                x, y, level = tile.coord
                if self.supports_timestamp:
                    records.append((level, x, y, content, time.time()))
                else:
                    records.append((level, x, y, content))

        cursor = self.db.cursor()
        try:
            if self.supports_timestamp:
                stmt = ("INSERT OR REPLACE INTO tiles (zoom_level, tile_column, tile_row, tile_data, last_modified)"
                        " VALUES (?,?,?,?, datetime(?, 'unixepoch', 'localtime'))")
                cursor.executemany(stmt, records)
            else:
                stmt = "INSERT OR REPLACE INTO tiles (zoom_level, tile_column, tile_row, tile_data) VALUES (?,?,?,?)"
                cursor.executemany(stmt, records)
            self.db.commit()
        except sqlite3.OperationalError as ex:
            log.warning('unable to store tile: %s', ex)
            return False
        return True

    def load_tile(self, tile, with_metadata=False, dimensions=None):
        if tile.source or tile.coord is None:
            return True

        cur = self.db.cursor()
        if self.supports_timestamp:
            stmt = '''SELECT tile_data, last_modified
                FROM tiles
                WHERE tile_column = ? AND
                      tile_row = ? AND
                      zoom_level = ?'''
        else:
            stmt = '''SELECT tile_data FROM tiles
                WHERE tile_column = ? AND
                      tile_row = ? AND
                      zoom_level = ?'''

        if self.ttl:
            stmt += " AND datetime('now', 'localtime', '%d seconds') < last_modified" % -self.ttl

        cur.execute(stmt, tile.coord)

        content = cur.fetchone()
        if content:
            tile.source = ImageSource(BytesIO(content[0]))
            if self.supports_timestamp:
                tile.timestamp = sqlite_datetime_to_timestamp(content[1])
            return True
        else:
            return False

    def load_tiles(self, tiles, with_metadata=False, dimensions=None):
        # associate the right tiles with the cursor
        tile_dict = {}
        coords = []
        for tile in tiles:
            if tile.source or tile.coord is None:
                continue
            x, y, level = tile.coord
            coords.append(x)
            coords.append(y)
            coords.append(level)
            tile_dict[(x, y)] = tile

        if not tile_dict:
            # all tiles loaded or coords are None
            return True

        if self.supports_timestamp:
            stmt_base = "SELECT tile_column, tile_row, tile_data, last_modified FROM tiles WHERE "
            if self.ttl:
                ttl_condition = "datetime('now', 'localtime', '%d seconds') < last_modified" % -self.ttl
                stmt_base += ttl_condition + ' AND '
        else:
            stmt_base = "SELECT tile_column, tile_row, tile_data FROM tiles WHERE "

        loaded_tiles = 0

        # SQLite is limited to 1000 args -> split into multiple requests if more arguments are needed
        while coords:
            cur_coords = coords[:999]

            stmt = stmt_base + '(' + ' OR '.join(
                ['(tile_column = ? AND tile_row = ? AND zoom_level = ?)'] * (len(cur_coords) // 3)) + ')'

            cursor = self.db.cursor()
            cursor.execute(stmt, cur_coords)

            for row in cursor:
                loaded_tiles += 1
                tile = tile_dict[(row[0], row[1])]
                data = row[2]
                tile.size = len(data)
                tile.source = ImageSource(BytesIO(data))
                if self.supports_timestamp:
                    tile.timestamp = sqlite_datetime_to_timestamp(row[3])
            cursor.close()

            coords = coords[999:]

        return loaded_tiles == len(tile_dict)

    def remove_tile(self, tile, dimensions=None):
        cursor = self.db.cursor()
        cursor.execute(
            "DELETE FROM tiles WHERE (tile_column = ? AND tile_row = ? AND zoom_level = ?)",
            tile.coord)
        self.db.commit()
        if cursor.rowcount:
            return True
        return False

    def remove_level_tiles_before(self, level, timestamp=None, remove_all=False):
        if remove_all:
            cursor = self.db.cursor()
            cursor.execute(
                "DELETE FROM tiles WHERE (zoom_level = ?)",
                (level, ))
            self.db.commit()
            if cursor.rowcount:
                return True
            return False

        if self.supports_timestamp:
            cursor = self.db.cursor()
            cursor.execute(
                "DELETE FROM tiles WHERE (zoom_level = ? AND last_modified < datetime(?, 'unixepoch', 'localtime'))",
                (level, timestamp))
            self.db.commit()
            if cursor.rowcount:
                return True
            return False

    def load_tile_metadata(self, tile, dimensions=None):
        if not self.supports_timestamp:
            # MBTiles specification does not include timestamps.
            # This sets the timestamp of the tile to epoch (1970s)
            tile.timestamp = -1
        else:
            self.load_tile(tile, dimensions=dimensions)


class MBTilesLevelCache(TileCacheBase):
    supports_timestamp = True

    def __init__(self, mbtiles_dir, timeout=30, wal=False, ttl=0, coverage=None,
                 directory_permissions=None, file_permissions=None):
        super(MBTilesLevelCache, self).__init__(coverage)
        md5 = hashlib.new('md5', mbtiles_dir.encode('utf-8'), usedforsecurity=False)
        self.lock_cache_id = 'sqlite-' + md5.hexdigest()
        self.cache_dir = mbtiles_dir
        self.directory_permissions = directory_permissions
        self.file_permissions = file_permissions
        self._mbtiles = {}
        self.timeout = timeout
        self.wal = wal
        self.ttl = ttl
        self._mbtiles_lock = threading.Lock()

    def _get_level(self, level):
        if level in self._mbtiles:
            return self._mbtiles[level]

        with self._mbtiles_lock:
            if level not in self._mbtiles:
                mbtile_filename = os.path.join(self.cache_dir, '%s.mbtiles' % level)
                self._mbtiles[level] = MBTilesCache(
                    mbtile_filename,
                    with_timestamps=True,
                    timeout=self.timeout,
                    wal=self.wal,
                    ttl=self.ttl,
                    coverage=self.coverage,
                    directory_permissions=self.directory_permissions,
                    file_permissions=self.file_permissions
                )

        return self._mbtiles[level]

    def cleanup(self):
        """
        Close all open connection and remove them from cache.
        """
        with self._mbtiles_lock:
            for mbtile in self._mbtiles.values():
                mbtile.cleanup()

    def is_cached(self, tile, dimensions=None):
        if tile.coord is None:
            return True
        if tile.source:
            return True

        return self._get_level(tile.coord[2]).is_cached(tile, dimensions=dimensions)

    def store_tile(self, tile, dimensions=None):
        if tile.stored:
            return True

        return self._get_level(tile.coord[2]).store_tile(tile, dimensions=dimensions)

    def store_tiles(self, tiles, dimensions=None):
        failed = False
        for level, tiles in groupby(tiles, key=lambda t: t.coord[2]):
            tiles = [t for t in tiles if not t.stored]
            res = self._get_level(level).store_tiles(tiles, dimensions=dimensions)
            if not res:
                failed = True
        return failed

    def load_tile(self, tile, with_metadata=False, dimensions=None):
        if tile.source or tile.coord is None:
            return True

        return self._get_level(tile.coord[2]).load_tile(tile, with_metadata=with_metadata, dimensions=dimensions)

    def load_tiles(self, tiles, with_metadata=False, dimensions=None):
        level = None
        for tile in tiles:
            if tile.source or tile.coord is None:
                continue
            level = tile.coord[2]
            break

        if not level:
            return True

        return self._get_level(level).load_tiles(tiles, with_metadata=with_metadata, dimensions=dimensions)

    def remove_tile(self, tile, dimensions=None):
        if tile.coord is None:
            return True

        return self._get_level(tile.coord[2]).remove_tile(tile)

    def load_tile_metadata(self, tile, dimensions=None):
        self.load_tile(tile, dimensions=dimensions)

    def remove_level_tiles_before(self, level, timestamp=None, remove_all=False):
        level_cache = self._get_level(level)
        if remove_all:
            level_cache.cleanup()
            os.unlink(level_cache.mbtile_file)
            for file in glob.glob("%s-*" % glob.escape(level_cache.mbtile_file)):
                os.unlink(file)
            return True
        else:
            return level_cache.remove_level_tiles_before(level, timestamp)



# This file is part of the MapProxy project.
# Copyright (C) 2010-2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from io import BytesIO
import pytest

from PIL import Image
from mapproxy.test.image import tmp_image
from mapproxy.test.http import mock_httpd
from mapproxy.test.system import SysTest
from mapproxy.util.times import timestamp_from_isodate


@pytest.fixture(scope="module")
def config_file():
    return "tileservice_refresh.yaml"


class TestRefresh(SysTest):

    def test_refresh_tile_1s(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=-20037508.3428,-20037508.3428,0.0,0.0"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tiles/wms_cache/1/0/0.jpeg")
                assert resp.content_type == "image/jpeg"
                file_path = cache_dir.join(
                    "wms_cache_EPSG900913/01/000/000/000/000/000/000.jpeg"
                )
                assert file_path.check()
                t1 = file_path.mtime()
                # file_path.remove()
                # assert not file_path.check()
            resp = app.get("/tiles/wms_cache/1/0/0.jpeg")
            assert resp.content_type == "image/jpeg"
            assert file_path.check()
            t2 = file_path.mtime()
            # tile is expired after 1 sec, so it will be fetched again from mock server
            time.sleep(1.2)
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tiles/wms_cache/1/0/0.jpeg")
                assert resp.content_type == "image/jpeg"
                assert file_path.check()
                t3 = file_path.mtime()
        assert t2 == t1
        assert t3 > t2

    def test_refresh_tile_mtime(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=-20037508.3428,-20037508.3428,0.0,0.0"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tiles/wms_cache_isotime/1/0/0.jpeg")
                assert resp.content_type == "image/jpeg"
                file_path = cache_dir.join(
                    "wms_cache_isotime_EPSG900913/01/000/000/000/000/000/000.jpeg"
                )
                assert file_path.check()
                timestamp = timestamp_from_isodate("2009-02-15T23:31:30")
                file_path.setmtime(timestamp + 1.2)
                t1 = file_path.mtime()
            resp = app.get("/tiles/wms_cache_isotime/1/0/0.jpeg")
            assert resp.content_type == "image/jpeg"
            t2 = file_path.mtime()
            file_path.setmtime(timestamp - 1.2)
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tiles/wms_cache_isotime/1/0/0.jpeg")
                assert resp.content_type == "image/jpeg"
                assert file_path.check()
                t3 = file_path.mtime()
            assert t2 == t1
            assert t3 > t2

    def test_refresh_tile_source_error_no_stale(self, app, cache_dir):
        source_request = {
            "path": r"/service?LAYERs=bar&SERVICE=WMS&FORMAT=image%2Fpng"
            "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
            "&VERSION=1.1.1&BBOX=-20037508.3428,-20037508.3428,0.0,0.0"
            "&WIDTH=256"
        }
        with tmp_image((256, 256), format="png") as img:
            expected_req = (
                source_request,
                {"body": img.read(), "headers": {"content-type": "image/png"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tiles/wms_cache_png/1/0/0.png")
                assert resp.content_type == "image/png"
                img.seek(0)
                assert resp.body == img.read()
            resp = app.get("/tiles/wms_cache_png/1/0/0.png")
            assert resp.content_type == "image/png"
            img.seek(0)
            assert resp.body == img.read()
            # tile is expired after 1 sec, so it will be requested again from mock server
            time.sleep(1.2)
            expected_req = (
                source_request,
                {"body": "", "status": 404},
            )
            with mock_httpd(
                    ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tiles/wms_cache_png/1/0/0.png")
                assert resp.content_type == "image/png"
                # error handler for 404 does not authorise stale tiles, so transparent tile will be rendered
                resp_img = Image.open(BytesIO(resp.body))
                # check response transparency
                assert resp_img.getbands() == ('R', 'G', 'B', 'A')
                assert resp_img.getextrema()[3] == (0, 0)

            expected_req = (
                source_request,
                {"body": "", "status": 405},
            )
            with mock_httpd(
                    ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tiles/wms_cache_png/1/0/0.png")
                assert resp.content_type == "image/png"
                # error handler for 405 does not authorise stale tiles, so red tile will be rendered
                resp_img = Image.open(BytesIO(resp.body))
                # check response red color
                assert resp_img.getbands() == ('R', 'G', 'B')
                assert resp_img.getextrema() == ((255, 255), (0, 0), (0, 0))

    def test_refresh_tile_source_error_stale(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=-20037508.3428,-20037508.3428,0.0,0.0"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tiles/wms_cache/1/0/0.jpeg")
                assert resp.content_type == "image/jpeg"
                img.seek(0)
                assert resp.body == img.read()
            resp = app.get("/tiles/wms_cache/1/0/0.jpeg")
            assert resp.content_type == "image/jpeg"
            img.seek(0)
            assert resp.body == img.read()
            # tile is expired after 1 sec, so it will be fetched again from mock server
            time.sleep(1.2)
            expected_req = (
                {
                    "path": r"/service?LAYERs=bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=-20037508.3428,-20037508.3428,0.0,0.0"
                    "&WIDTH=256"
                },
                {"body": "", "status": 406},
            )
            with mock_httpd(
                    ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tiles/wms_cache/1/0/0.jpeg")
                assert resp.content_type == "image/jpeg"
                # Check that initial non empty img is served as a stale tile
                img.seek(0)
                assert resp.body == img.read()



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

from io import BytesIO

import pytest

from PIL import Image
from mapproxy.test.system import SysTest


has_mapnik = True
try:
    import mapnik
except ImportError:
    try:
        import mapnik2 as mapnik  # noqa
    except ImportError:
        has_mapnik = False


mapnik_xml = (
    b"""
<?xml version="1.0"?>
<!DOCTYPE Map>
<Map background-color="#ff0000" bgcolor="#ff0000" srs="+proj=latlong +datum=WGS84">
    <Layer name="marker">
        <StyleName>marker</StyleName>
        <Datasource>
            <Parameter name="type">ogr</Parameter>
            <Parameter name="file">test_point.geojson</Parameter>
            <Parameter name="layer">OGRGeoJSON</Parameter>
        </Datasource>
    </Layer>
    <Style name="marker">
        <Rule>
            <MarkersSymbolizer fill="transparent" width="10" height="10" stroke="black" stroke-width="3"
                placement="point" marker-type="ellipse"/>
        </Rule>
    </Style>
</Map>
""".strip()
)

test_point_geojson = (
    b"""
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [-45, -45]}, "properties": {}}
""".strip()
)

mapnik_transp_xml = (
    b"""
<?xml version="1.0"?>
<!DOCTYPE Map>
<Map background-color="transparent" srs="+proj=latlong +datum=WGS84">
</Map>
""".strip()
)


@pytest.mark.skipif(not has_mapnik, reason="requires mapnik")
class TestMapnikSource(SysTest):

    @pytest.fixture(scope="class")
    def config_file(self):
        return "mapnik_source.yaml"

    @pytest.fixture(scope="class")
    def additional_files(self, base_dir):
        base_dir.join("test_point.geojson").write_binary(test_point_geojson)
        base_dir.join("mapnik.xml").write_binary(mapnik_xml)
        base_dir.join("mapnik-transparent.xml").write_binary(mapnik_transp_xml)

    def test_get_map(self, app):
        req = (
            r"/service?LAYERs=mapnik&SERVICE=WMS&FORMAT=image%2Fpng"
            "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A4326"
            "&VERSION=1.1.1&BBOX=-90,-90,0,0&styles="
            "&WIDTH=200&"
        )

        resp = app.get(req)
        data = BytesIO(resp.body)
        img = Image.open(data)
        colors = sorted(img.getcolors(), reverse=True)
        # map bg color + black marker
        assert 39700 < colors[0][0] < 39900, colors[0][0]
        assert colors[0][1] == (255, 0, 0, 255)
        assert 50 < colors[1][0] < 150, colors[1][0]
        assert colors[1][1] == (0, 0, 0, 255)

    def test_get_map_hq(self, app):
        req = (
            r"/service?LAYERs=mapnik_hq&SERVICE=WMS&FORMAT=image%2Fpng"
            "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A4326"
            "&VERSION=1.1.1&BBOX=-90,-90,0,0&styles="
            "&WIDTH=200&"
        )

        resp = app.get(req)
        data = BytesIO(resp.body)
        img = Image.open(data)
        colors = sorted(img.getcolors(), reverse=True)
        # map bg color + black marker (like above, but marker is scaled up)
        assert 39500 < colors[0][0] < 39600, colors[0][0]
        assert colors[0][1] == (255, 0, 0, 255)
        assert 250 < colors[1][0] < 350, colors[1][0]
        assert colors[1][1] == (0, 0, 0, 255)

    def test_get_map_outside_coverage(self, app):
        req = (
            r"/service?LAYERs=mapnik&SERVICE=WMS&FORMAT=image%2Fpng"
            "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A4326"
            "&VERSION=1.1.1&BBOX=-175,-85,-172,-82&styles="
            "&WIDTH=200&&BGCOLOR=0x00ff00"
        )

        resp = app.get(req)
        data = BytesIO(resp.body)
        img = Image.open(data)
        colors = sorted(img.getcolors(), reverse=True)
        # wms request bg color
        assert colors[0] == (40000, (0, 255, 0))

    def test_get_map_unknown_file(self, app):
        req = (
            r"/service?LAYERs=mapnik_unknown&SERVICE=WMS&FORMAT=image%2Fpng"
            "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A4326"
            "&VERSION=1.1.1&BBOX=-90,-90,0,0&styles="
            "&WIDTH=200&&BGCOLOR=0x00ff00"
        )

        resp = app.get(req)
        assert "unknown.xml" in resp.body, resp.body

    def test_get_map_transparent(self, app):
        req = (
            r"/service?LAYERs=mapnik_transparent&SERVICE=WMS&FORMAT=image%2Fpng"
            "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A4326"
            "&VERSION=1.1.1&BBOX=-90,-90,0,0&styles="
            "&WIDTH=200&transparent=True"
        )

        resp = app.get(req)
        data = BytesIO(resp.body)
        img = Image.open(data)
        colors = sorted(img.getcolors(), reverse=True)
        assert colors[0] == (40000, (0, 0, 0, 0))



import datetime
import sqlite3


def adapt_date_iso(val):
    """Adapt datetime.date to ISO 8601 date."""
    return val.isoformat()


def adapt_datetime_iso(val):
    """Adapt datetime.datetime to timezone-naive ISO 8601 date."""
    return val.isoformat()


def convert_date(val):
    """Convert ISO 8601 date to datetime.date object."""
    return datetime.date.fromisoformat(val)


def convert_datetime(val):
    """Convert ISO 8601 datetime to datetime.datetime object."""
    return datetime.datetime.fromisoformat(val)


sqlite3.register_adapter(datetime.date, adapt_date_iso)
sqlite3.register_adapter(datetime.datetime, adapt_datetime_iso)
sqlite3.register_converter('date', convert_date)
sqlite3.register_converter('datetime', convert_datetime)



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
System-wide configuration.
"""
import os
import copy
import contextlib
from mapproxy.util.yaml import load_yaml_file
from mapproxy.util.ext.local import LocalStack


class Options(dict):
    """
    Dictionary with attribute style access.

    >>> o = Options(bar='foo')
    >>> o.bar
    'foo'
    """

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, dict.__repr__(self))

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError(name)

    def update(self, other=None, **kw):
        if other is not None:
            if hasattr(other, 'iteritems'):
                it = other.iteritems()
            elif hasattr(other, 'items'):
                it = other.items()
            else:
                it = iter(other)
        else:
            it = iter(kw)
        for key, value in it:
            if key in self and isinstance(self[key], Options):
                self[key].update(value)
            else:
                self[key] = value

    def __deepcopy__(self, memo):
        return Options(copy.deepcopy(list(self.items()), memo))


_config = LocalStack()


def base_config():
    """
    Returns the context-local system-wide configuration.
    """
    config = _config.top
    if config is None:
        import warnings
        warnings.warn("calling un-configured base_config",
                      DeprecationWarning, stacklevel=2)
        config = load_default_config()
        config.conf_base_dir = os.getcwd()
        finish_base_config(config)
        _config.push(config)
    return config


@contextlib.contextmanager
def local_base_config(conf):
    """
    Temporarily set the global configuration (mapproxy.config.base_config).

    The global mapproxy.config.base_config object is thread-local and
    is set per-request in the MapProxyApp. Use `local_base_config` to
    set base_config outside of a request context (e.g. system loading
    or seeding).
    """
    import mapproxy.config.config
    mapproxy.config.config._config.push(conf)
    try:
        yield
    finally:
        mapproxy.config.config._config.pop()


def _to_options_map(mapping):
    if isinstance(mapping, dict):
        opt = Options()
        for key, value in mapping.items():
            opt[key] = _to_options_map(value)
        return opt
    elif isinstance(mapping, list):
        return [_to_options_map(m) for m in mapping]
    else:
        return mapping


def abspath(path, base_path=None):
    """
    Convert path to absolute path. Uses ``conf_base_dir`` as base, if
    path is relative and ``base_path`` is not set.
    """
    if base_path:
        return os.path.abspath(os.path.join(base_path, path))
    return os.path.join(base_config().conf_base_dir, path)


def finish_base_config(bc=None):
    bc = bc or base_config()
    if 'srs' in bc:
        # build union of default axis_order_xx_ and the user configured axis_order_xx
        default_ne = bc.srs.axis_order_ne_
        default_en = bc.srs.axis_order_en_
        # remove from default to allow overwrites
        default_ne.difference_update(set(bc.srs.axis_order_en))
        default_en.difference_update(set(bc.srs.axis_order_ne))
        bc.srs.axis_order_ne = default_ne.union(set(bc.srs.axis_order_ne))
        bc.srs.axis_order_en = default_en.union(set(bc.srs.axis_order_en))
        if 'proj_data_dir' in bc.srs:
            bc.srs.proj_data_dir = os.path.join(bc.conf_base_dir, bc.srs.proj_data_dir)

    if 'wms' in bc:
        bc.wms.srs = set(bc.wms.srs)

    if 'conf_base_dir' in bc:
        if 'cache' in bc:
            if 'base_dir' in bc.cache:
                bc.cache.base_dir = os.path.join(bc.conf_base_dir, bc.cache.base_dir)
            if 'lock_dir' in bc.cache:
                bc.cache.lock_dir = os.path.join(bc.conf_base_dir, bc.cache.lock_dir)


def load_base_config(config_file=None, clear_existing=False):
    """
    Load system wide base configuration.

    :param config_file: the file name of the mapproxy.yaml configuration.
                        if ``None``, load the internal proxylib/default.yaml conf
    :param clear_existing: if ``True`` remove the existing configuration settings,
                           else overwrite the settings.
    """

    if config_file is None:
        from mapproxy.config import defaults
        config_dict = {}
        for k, v in defaults.__dict__.items():
            if k.startswith('_'):
                continue
            config_dict[k] = v
        conf_base_dir = os.getcwd()
        load_config(base_config(), config_dict=config_dict, clear_existing=clear_existing)
    else:
        conf_base_dir = os.path.abspath(os.path.dirname(config_file))
        load_config(base_config(), config_file=config_file, clear_existing=clear_existing)

    bc = base_config()
    finish_base_config(bc)

    bc.conf_base_dir = conf_base_dir


def load_default_config():
    from mapproxy.config import defaults
    config_dict = {}
    for k, v in defaults.__dict__.items():
        if k.startswith('_'):
            continue
        config_dict[k] = v

    default_conf = Options()
    load_config(default_conf, config_dict=config_dict)
    return default_conf


def load_config(config, config_file=None, config_dict=None, clear_existing=False):
    if clear_existing:
        for key in list(config.keys()):
            del config[key]

    if config_dict is None:
        config_dict = load_yaml_file(config_file)

    defaults = _to_options_map(config_dict)

    if defaults:
        for key, value in defaults.items():
            if key in config and hasattr(config[key], 'update'):
                config[key].update(value)
            else:
                config[key] = value



# This file is part of the MapProxy project.
# Copyright (C) 2017 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time

import pytest

try:
    import redis
except ImportError:
    redis = None

from mapproxy.cache.redis import RedisCache
from mapproxy.cache.tile import Tile
from mapproxy.test.unit.test_cache_tile import TileCacheTestBase


@pytest.mark.skipif(not redis or not os.environ.get('MAPPROXY_TEST_REDIS'),
                    reason="redis package and MAPPROXY_TEST_REDIS env required")
class TestRedisCache(TileCacheTestBase):
    always_loads_metadata = False

    def setup_method(self):
        redis_host = os.environ['MAPPROXY_TEST_REDIS']
        self.host, self.port = redis_host.split(':')
        if os.environ.get('MAPPROXY_TEST_REDIS_TLS'):
            redis_host_tls = os.environ['MAPPROXY_TEST_REDIS_TLS']
            self.tls_host, self.tls_port = redis_host_tls.split(':')
        if os.environ.get('MAPPROXY_TEST_REDIS_AUTH'):
            redis_host_tls = os.environ['MAPPROXY_TEST_REDIS_AUTH']
            self.auth_host, self.auth_port = redis_host_tls.split(':')

        TileCacheTestBase.setup_method(self)

        self.cache = RedisCache(self.host, int(self.port), prefix='mapproxy-test', db=1)

    def teardown_method(self):
        for k in self.cache.r.keys('mapproxy-test-*'):
            self.cache.r.delete(k)

    def test_default_coverage(self):
        assert self.cache.coverage is None

    def test_expire(self):
        cache = RedisCache(self.host, int(self.port), prefix='mapproxy-test', db=1, ttl=0)
        t1 = self.create_tile(coord=(9382, 1234, 9))
        assert cache.store_tile(t1)
        time.sleep(0.1)
        t2 = Tile(t1.coord)
        assert cache.is_cached(t2)

        cache = RedisCache(self.host, int(self.port), prefix='mapproxy-test', db=1, ttl=0.05)
        t1 = self.create_tile(coord=(5382, 2234, 9))
        assert cache.store_tile(t1)
        time.sleep(0.1)
        t2 = Tile(t1.coord)
        assert not cache.is_cached(t2)

    def test_double_remove(self):
        tile = self.create_tile()
        self.create_cached_tile(tile)
        assert self.cache.remove_tile(tile)
        assert self.cache.remove_tile(tile)

    @pytest.mark.skipif(not redis or not os.environ.get('MAPPROXY_TEST_REDIS_TLS'),
                        reason="MAPPROXY_TEST_REDIS_TLS is required")
    def test_tls_authentication_enabled(self):
        print(os.curdir)
        ssl_certfile = 'mapproxy/test/unit/fixture/redis-client.crt'
        ssl_keyfile = 'mapproxy/test/unit/fixture/redis-client.key'
        ssl_ca_certs = 'mapproxy/test/unit/fixture/ca.crt'
        cache = RedisCache(
            self.tls_host, int(self.tls_port), prefix='mapproxy-test', db=1, ssl_certfile=ssl_certfile,
            ssl_keyfile=ssl_keyfile, ssl_ca_certs=ssl_ca_certs)
        assert cache.r.connection_pool.connection_kwargs['ssl_certfile'] == ssl_certfile
        assert cache.r.connection_pool.connection_kwargs['ssl_keyfile'] == ssl_keyfile
        assert cache.r.connection_pool.connection_kwargs['ssl_ca_certs'] == ssl_ca_certs
        t1 = self.create_tile(coord=(5382, 3234, 9))
        assert cache.store_tile(t1)
        time.sleep(0.1)
        t2 = Tile(t1.coord)
        assert cache.is_cached(t2)

    @pytest.mark.skipif(not redis or not os.environ.get('MAPPROXY_TEST_REDIS_TLS'),
                        reason="MAPPROXY_TEST_REDIS_TLS is required")
    def test_tls_authentication_disabled(self):
        cache = RedisCache(self.tls_host, int(self.tls_port), prefix='mapproxy-test', db=1)
        assert 'ssl_certfile' not in cache.r.connection_pool.connection_kwargs
        assert 'ssl_keyfile' not in cache.r.connection_pool.connection_kwargs
        assert 'ssl_ca_certs' not in cache.r.connection_pool.connection_kwargs
        assert not cache.r.connection_pool.connection_kwargs.get('ssl', False)
        t1 = self.create_tile(coord=(5382, 4234, 9))
        assert not cache.store_tile(t1)
        time.sleep(0.1)
        t2 = Tile(t1.coord)
        assert not cache.is_cached(t2)

    @pytest.mark.skipif(not redis or not os.environ.get('MAPPROXY_TEST_REDIS_AUTH'),
                        reason="MAPPROXY_TEST_REDIS_AUTH is required to test authentication")
    def test_user_password_authentication(self):
        username = 'test'
        password = 'pw4test'
        cache = RedisCache(
            self.auth_host, int(self.auth_port), prefix='mapproxy-test', db=1, username=username, password=password)
        assert cache.r.connection_pool.connection_kwargs['username'] == username
        assert cache.r.connection_pool.connection_kwargs['password'] == password
        t1 = self.create_tile(coord=(5382, 5234, 9))
        assert cache.store_tile(t1)
        t2 = Tile(t1.coord)
        assert cache.is_cached(t2)



import os

from ..parse import parse_capabilities


def local_filename(filename):
    return os.path.join(os.path.dirname(__file__), filename)


class TestWMS111(object):

    def test_parse_metadata(self):
        cap = parse_capabilities(local_filename("wms-example-111.xml"))
        md = cap.metadata()
        assert md["name"] == "OGC:WMS"
        assert md["title"] == "ACME OpenStreetMap WMS"
        assert md["access_constraints"] == "Here be dragons."
        assert md["fees"] == "none"
        assert md["online_resource"] == "http://example.org/"
        assert md["abstract"] == "ACME OpenStreetMap WMS (powered by MapProxy)"

        assert md["contact"]["person"] == "Bob Mustermann"
        assert md["contact"]["organization"] == "ACME"
        assert md["contact"]["position"] == "Technical Director"
        assert md["contact"]["address"] == "Fakestreet 123"
        assert md["contact"]["city"] == "Fakecity"
        assert md["contact"]["postcode"] == "12345"
        assert md["contact"]["country"] == "Germany"
        assert md["contact"]["phone"] == "0123456789"
        assert md["contact"]["fax"] == "0123456789"
        assert md["contact"]["email"] == "info@example.org"

    def test_parse_layer(self):
        cap = parse_capabilities(local_filename("wms-example-111.xml"))
        lyrs = cap.layers_list()
        assert len(lyrs) == 2
        assert lyrs[0]["llbbox"] == [-180.0, -85.0511287798, 180.0, 85.0511287798]
        assert lyrs[0]["srs"] == {"EPSG:4326", "EPSG:4258", "CRS:84", "EPSG:900913", "EPSG:31466", "EPSG:31467",
                                  "EPSG:31468", "EPSG:25831", "EPSG:25832", "EPSG:25833", "EPSG:3857"}
        assert len(lyrs[0]["bbox_srs"]) == 1
        assert lyrs[0]["bbox_srs"]["EPSG:4326"] == [
            -180.0,
            -85.0511287798,
            180.0,
            85.0511287798,
        ]

    def test_parse_layer_2(self):
        cap = parse_capabilities(local_filename("wms-large-111.xml"))
        lyrs = cap.layers_list()
        assert len(lyrs) == 46
        assert lyrs[0]["llbbox"] == [-10.4, 35.7, 43.0, 74.1]
        assert lyrs[0]["srs"] == {"EPSG:31467", "EPSG:31466", "EPSG:31465", "EPSG:31464", "EPSG:31463", "EPSG:31462",
                                  "EPSG:4326", "EPSG:31469", "EPSG:31468", "EPSG:31257", "EPSG:31287", "EPSG:31286",
                                  "EPSG:31285", "EPSG:31284", "EPSG:31258", "EPSG:31259", "EPSG:31492", "EPSG:31493",
                                  "EPSG:25833", "EPSG:25832", "EPSG:31494", "EPSG:31495", "EPSG:28992"}
        assert lyrs[1]["name"] == "Grenzen"
        assert (
            lyrs[1]["legend"]["url"]
            == "http://example.org/service?SERVICE=WMS&version=1.1.1&service=WMS&request=GetLegendGraphic&layer=Grenzen&format=image/png&STYLE=default"  # noqa
        )


class TestWMS130(object):

    def test_parse_metadata(self):
        cap = parse_capabilities(local_filename("wms-example-130.xml"))
        md = cap.metadata()
        assert md["name"] == "WMS"
        assert md["title"] == "ACME OpenStreetMap WMS"

        req = cap.requests()
        assert req["GetMap"] == "http://example.org/service"

    def test_parse_layer(self):
        cap = parse_capabilities(local_filename("wms-example-130.xml"))
        lyrs = cap.layers_list()
        assert len(lyrs) == 2
        assert lyrs[0]["llbbox"] == [-180.0, -85.0511287798, 180.0, 85.0511287798]
        assert lyrs[0]["srs"] == {"EPSG:4326", "EPSG:4258", "CRS:84", "EPSG:900913", "EPSG:31466", "EPSG:31467",
                                  "EPSG:31468", "EPSG:25831", "EPSG:25832", "EPSG:25833", "EPSG:3857"}
        assert len(lyrs[0]["bbox_srs"]) == 4
        assert set(lyrs[0]["bbox_srs"].keys()) == {"CRS:84", "EPSG:900913", "EPSG:4326", "EPSG:3857"}
        assert lyrs[0]["bbox_srs"]["EPSG:3857"] == [
            -20037508.3428,
            -20037508.3428,
            20037508.3428,
            20037508.3428,
        ]
        # EPSG:4326 bbox should be switched to long/lat
        assert lyrs[0]["bbox_srs"]["EPSG:4326"] == (
            -180.0,
            -85.0511287798,
            180.0,
            85.0511287798,
        )


class TestLargeWMSCapabilities(object):

    def test_parse_metadata(self):
        cap = parse_capabilities(local_filename("wms_nasa_cap.xml"))
        md = cap.metadata()
        assert md["name"] == "OGC:WMS"
        assert md["title"] == "JPL Global Imagery Service"

    def test_parse_layer(self):
        cap = parse_capabilities(local_filename("wms_nasa_cap.xml"))
        lyrs = cap.layers_list()
        assert len(lyrs) == 15
        assert len(lyrs[0]["bbox_srs"]) == 0



# This file is part of the MapProxy project.
# Copyright (C) 2025 terrestris GmbH & Co. KG <https://terrestris.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from mapproxy.srs import SRS
from mapproxy.util.bbox import bbox_tuple


class GridError(Exception):
    pass


class NoTiles(GridError):
    pass


def _create_tile_list(xs, ys, level, grid_size):
    """
    Returns an iterator tile_coords for the given tile ranges (`xs` and `ys`).
    If the one tile_coord is negative or out of the `grid_size` bound,
    the coord is None.
    """
    x_limit = grid_size[0]
    y_limit = grid_size[1]
    for y in ys:
        for x in xs:
            if x < 0 or y < 0 or x >= x_limit or y >= y_limit:
                yield None
            else:
                yield x, y, level


class _default_bboxs(object):
    _defaults = {
        4326: (-180, -90, 180, 90),
    }
    for epsg_num in (900913, 3857, 102100, 102113):
        _defaults[epsg_num] = (-20037508.342789244,
                               -20037508.342789244,
                               20037508.342789244,
                               20037508.342789244)
    defaults = None

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key):
        if self.defaults is None:
            defaults = {}
            for epsg, bbox in self._defaults.items():
                defaults[SRS(epsg)] = bbox
            self.defaults = defaults
        return self.defaults[key]


default_bboxs = _default_bboxs()


def grid_bbox(bbox, bbox_srs, srs):
    bbox = bbox_tuple(bbox)
    if bbox_srs:
        bbox = SRS(bbox_srs).transform_bbox_to(srs, bbox)
    return bbox


ORIGIN_UL = 'ul'
ORIGIN_LL = 'll'


def origin_from_string(origin):
    if origin is None:
        origin = ORIGIN_LL
    elif origin.lower() in ('ll', 'sw'):
        origin = ORIGIN_LL
    elif origin.lower() in ('ul', 'nw'):
        origin = ORIGIN_UL
    else:
        raise ValueError("unknown origin value '%s'" % origin)
    return origin



# This file is part of the MapProxy project.
# Copyright (C) 2010-2016 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from mapproxy.util.fs import ensure_directory
from mapproxy.request.base import NoCaseMultiDict


def location_funcs(layout):
    if layout == 'tc':
        return tile_location_tc, level_location
    elif layout == 'mp':
        return tile_location_mp, level_location
    elif layout == 'tms':
        return tile_location_tms, level_location
    elif layout == 'reverse_tms':
        return tile_location_reverse_tms, None
    elif layout == 'quadkey':
        return tile_location_quadkey, no_level_location
    elif layout == 'arcgis':
        return tile_location_arcgiscache, level_location_arcgiscache
    else:
        raise ValueError('unknown directory_layout "%s"' % layout)


def level_location(level, cache_dir, dimensions=None):
    """
    Return the path where all tiles for `level` will be stored.

    >>> os.path.abspath(level_location(2, '/tmp/cache')) == os.path.abspath('/tmp/cache/02')
    True
    """
    dim_path = dimensions_part(dimensions)

    if isinstance(level, str):
        return os.path.join(cache_dir, dim_path, level)
    else:
        return os.path.join(cache_dir, dim_path, "%02d" % level)


def dimensions_part(dimensions):
    """
    Return the subpath where all tiles for `dimensions` will be stored.
    Dimensions prefixed with "dim_" are sorted after the predefined elevation and time dimensions
    >>> dimensions_part({'time': '2020-08-25T00:00:00Z'})
    'time-2020-08-25T00:00:00Z'
    >>> dimensions_part({'time': '2020-08-25T00:00:00Z', 'dim_reference_time': '2020-08-25T00:00:00Z', 'dim_level': '700'})  # noqa
    'time-2020-08-25T00:00:00Z/dim_level-700/dim_reference_time-2020-08-25T00:00:00Z'

    """
    if dimensions:
        dims = NoCaseMultiDict(dimensions)
        predefined_dims, custom_dims = [], []
        for dim in dims.keys():
            (custom_dims if dim.startswith('dim_') else predefined_dims).append(dim)
        dim_keys = sorted(predefined_dims) + sorted(custom_dims)
        return os.path.join(*(map(lambda k: k + "-" + str(dims.get(k, 'default')), dim_keys)))
    else:
        return ""


def level_part(level):
    """
    Return the path where all tiles for `level` will be stored.

    >>> level_part(2)
    '02'
    >>> level_part('2')
    '2'
    """
    if isinstance(level, str):
        return level
    else:
        return "%02d" % level


def tile_location_tc(tile, cache_dir, file_ext, create_dir=False, dimensions=None,
                     directory_permissions=None):
    """
    Return the location of the `tile`. Caches the result as ``location``
    property of the `tile`.

    :param tile: the tile object
    :param create_dir: if True, create all necessary directories
    :return: the full filename of the tile

    >>> from mapproxy.cache.tile import Tile
    >>> tile_location_tc(Tile((3, 4, 2)), '/tmp/cache', 'png').replace('\\\\', '/')
    '/tmp/cache/02/000/000/003/000/000/004.png'
    """
    if tile.location is None:
        x, y, z = tile.coord

        parts = (cache_dir,
                 dimensions_part(dimensions),
                 level_part(z),
                 "%03d" % int(x / 1000000),
                 "%03d" % (int(x / 1000) % 1000),
                 "%03d" % (int(x) % 1000),
                 "%03d" % int(y / 1000000),
                 "%03d" % (int(y / 1000) % 1000),
                 "%03d.%s" % (int(y) % 1000, file_ext))
        tile.location = os.path.join(*parts)
    if create_dir:
        ensure_directory(tile.location, directory_permissions)
    return tile.location


def tile_location_mp(tile, cache_dir, file_ext, create_dir=False, dimensions=None,
                     directory_permissions=None):
    """
    Return the location of the `tile`. Caches the result as ``location``
    property of the `tile`.

    :param tile: the tile object
    :param create_dir: if True, create all necessary directories
    :return: the full filename of the tile

    >>> from mapproxy.cache.tile import Tile
    >>> tile_location_mp(Tile((3, 4, 2)), '/tmp/cache', 'png').replace('\\\\', '/')
    '/tmp/cache/02/0000/0003/0000/0004.png'
    >>> tile_location_mp(Tile((12345678, 98765432, 22)), '/tmp/cache', 'png').replace('\\\\', '/')
    '/tmp/cache/22/1234/5678/9876/5432.png'
    """
    if tile.location is None:
        x, y, z = tile.coord
        parts = (cache_dir,
                 dimensions_part(dimensions),
                 level_part(z),
                 "%04d" % int(x / 10000),
                 "%04d" % (int(x) % 10000),
                 "%04d" % int(y / 10000),
                 "%04d.%s" % (int(y) % 10000, file_ext))
        tile.location = os.path.join(*parts)
    if create_dir:
        ensure_directory(tile.location, directory_permissions)
    return tile.location


def tile_location_tms(tile, cache_dir, file_ext, create_dir=False, dimensions=None,
                      directory_permissions=None):
    """
    Return the location of the `tile`. Caches the result as ``location``
    property of the `tile`.

    :param tile: the tile object
    :param create_dir: if True, create all necessary directories
    :return: the full filename of the tile

    >>> from mapproxy.cache.tile import Tile
    >>> tile_location_tms(Tile((3, 4, 2)), '/tmp/cache', 'png').replace('\\\\', '/')
    '/tmp/cache/2/3/4.png'
    """
    if tile.location is None:
        x, y, z = tile.coord
        tile.location = os.path.join(
            cache_dir, dimensions_part(dimensions), level_part(str(z)),
            str(x), str(y) + '.' + file_ext
        )
    if create_dir:
        ensure_directory(tile.location, directory_permissions)
    return tile.location


def tile_location_reverse_tms(tile, cache_dir, file_ext, create_dir=False, dimensions=None,
                              directory_permissions=None):
    """
    Return the location of the `tile`. Caches the result as ``location``
    property of the `tile`.

    :param tile: the tile object
    :param create_dir: if True, create all necessary directories
    :return: the full filename of the tile

    >>> from mapproxy.cache.tile import Tile
    >>> tile_location_reverse_tms(Tile((3, 4, 2)), '/tmp/cache', 'png').replace('\\\\', '/')
    '/tmp/cache/4/3/2.png'
    """
    if tile.location is None:
        x, y, z = tile.coord
        tile.location = os.path.join(
            cache_dir, dimensions_part(dimensions), str(y), str(x), str(z) + '.' + file_ext
        )
    if create_dir:
        ensure_directory(tile.location, directory_permissions)
    return tile.location


def level_location_tms(level, cache_dir, dimensions=None):
    return level_location(str(level), cache_dir=cache_dir)


def tile_location_quadkey(tile, cache_dir, file_ext, create_dir=False, dimensions=None,
                          directory_permissions=None):
    """
    Return the location of the `tile`. Caches the result as ``location``
    property of the `tile`.

    :param tile: the tile object
    :param create_dir: if True, create all necessary directories
    :return: the full filename of the tile

    >>> from mapproxy.cache.tile import Tile
    >>> tile_location_quadkey(Tile((3, 4, 2)), '/tmp/cache', 'png').replace('\\\\', '/')
    '/tmp/cache/11.png'
    """
    if tile.location is None:
        x, y, z = tile.coord
        quadKey = ""
        for i in range(z, 0, -1):
            digit = 0
            mask = 1 << (i-1)
            if (x & mask) != 0:
                digit += 1
            if (y & mask) != 0:
                digit += 2
            quadKey += str(digit)
        tile.location = os.path.join(
            cache_dir, quadKey + '.' + file_ext
        )
    if create_dir:
        ensure_directory(tile.location, directory_permissions)
    return tile.location


def no_level_location(level, cache_dir, dimensions=None):
    # dummy for quadkey cache which stores all tiles in one directory
    raise NotImplementedError('cache does not have any level location')


def tile_location_arcgiscache(tile, cache_dir, file_ext, create_dir=False, dimensions=None,
                              directory_permissions=None):
    """
    Return the location of the `tile`. Caches the result as ``location``
    property of the `tile`.

    :param tile: the tile object
    :param create_dir: if True, create all necessary directories
    :return: the full filename of the tile

    >>> from mapproxy.cache.tile import Tile
    >>> tile_location_arcgiscache(Tile((1234567, 87654321, 9)), '/tmp/cache', 'png').replace('\\\\', '/')
    '/tmp/cache/L09/R05397fb1/C0012d687.png'
    """
    if tile.location is None:
        x, y, z = tile.coord
        parts = (cache_dir, 'L%02d' % z, 'R%08x' % y, 'C%08x.%s' % (x, file_ext))
        tile.location = os.path.join(*parts)
    if create_dir:
        ensure_directory(tile.location, directory_permissions)
    return tile.location


def level_location_arcgiscache(z, cache_dir, dimensions=None):
    return level_location('L%02d' % z, cache_dir=cache_dir, dimensions=dimensions)



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import logging

import os
import time
import operator
from functools import reduce

from mapproxy.cache.dummy import DummyCache
from mapproxy.config import abspath
from mapproxy.config.coverage import load_coverage
from mapproxy.config.loader import ConfigurationError
from mapproxy.seed.util import bidict
from mapproxy.seed.seeder import SeedTask, CleanupTask
from mapproxy.seed.spec import validate_seed_conf
from mapproxy.util.bbox import TransformationError
from mapproxy.util.coverage import MultiCoverage, BBOXCoverage
from mapproxy.util.geom import GeometryError, EmptyGeometryError, CoverageReadError
from mapproxy.util.py import memoize
from mapproxy.util.times import timestamp_from_isodate, timestamp_before
from mapproxy.util.yaml import load_yaml_file, YAMLError


class SeedConfigurationError(ConfigurationError):
    pass


class EmptyCoverageError(Exception):
    pass


log = logging.getLogger('mapproxy.seed.config')


def load_seed_tasks_conf(seed_conf_filename, mapproxy_conf):
    try:
        conf = load_yaml_file(seed_conf_filename)
    except YAMLError as ex:
        raise SeedConfigurationError(ex)

    if 'views' in conf:
        raise Exception('The old seeding config style is no longer supported. Please refer to the documentation.')
    else:
        errors, informal_only = validate_seed_conf(conf)
        for error in errors:
            log.warning(error)
        if not informal_only:
            raise SeedConfigurationError('invalid configuration')
        seed_conf = SeedingConfiguration(conf, mapproxy_conf=mapproxy_conf)
    return seed_conf


class SeedingConfiguration(object):
    def __init__(self, seed_conf, mapproxy_conf):
        self.conf = seed_conf
        self.mapproxy_conf = mapproxy_conf
        self.grids = bidict((name, grid_conf.tile_grid()) for name, grid_conf in self.mapproxy_conf.grids.items())

    @memoize
    def coverage(self, name):
        coverage_conf = (self.conf.get('coverages') or {}).get(name)
        if coverage_conf is None:
            raise SeedConfigurationError('coverage %s not found. available coverages: %s' % (
                name, ','.join((self.conf.get('coverages') or {}).keys())))

        try:
            coverage = load_coverage(coverage_conf)
        except CoverageReadError as ex:
            raise SeedConfigurationError("can't load coverage '%s'. %s" % (name, ex))
        except GeometryError as ex:
            raise SeedConfigurationError("invalid geometry in coverage '%s'. %s" % (name, ex))
        except EmptyGeometryError as ex:
            raise EmptyCoverageError("coverage '%s' contains no geometries. %s" % (name, ex))
        except Exception:
            raise Exception(f"can't load coverage '{name}'")

        # without extend we have an empty coverage
        if not coverage.extent.llbbox:
            raise EmptyCoverageError("coverage '%s' contains no geometries." % name)
        return coverage

    def cache(self, cache_name):
        cache = {}
        if cache_name not in self.mapproxy_conf.caches:
            raise SeedConfigurationError('cache %s not found. available caches: %s' % (
                cache_name, ','.join(self.mapproxy_conf.caches.keys())))
        for tile_grid, extent, tile_mgr in self.mapproxy_conf.caches[cache_name].caches():
            if isinstance(tile_mgr.cache, DummyCache):
                raise SeedConfigurationError('can\'t seed cache %s (disable_storage: true)' %
                                             cache_name)
            grid_name = self.grids[tile_grid]
            cache[grid_name] = tile_mgr
        return cache

    def seed_tasks_names(self):
        seeds = self.conf.get('seeds') or {}
        return list(seeds.keys())

    def cleanup_tasks_names(self):
        cleanups = self.conf.get('cleanups') or {}
        return list(cleanups.keys())

    def seeds(self, names=None):
        """
        Return seed tasks.
        """
        tasks = []
        if names is None:
            names = (self.conf.get('seeds') or {}).keys()

        for seed_name in names:
            seed_conf = self.conf['seeds'][seed_name]
            seed_conf = SeedConfiguration(seed_name, seed_conf, self)
            for task in seed_conf.seed_tasks():
                tasks.append(task)
        return tasks

    def cleanups(self, names=None):
        """
        Return cleanup tasks.
        """
        tasks = []
        if names is None:
            names = (self.conf.get('cleanups') or {}).keys()

        for cleanup_name in names:
            cleanup_conf = self.conf['cleanups'][cleanup_name]
            cleanup_conf = CleanupConfiguration(cleanup_name, cleanup_conf, self)
            for task in cleanup_conf.cleanup_tasks():
                tasks.append(task)
        return tasks


class ConfigurationBase(object):
    def __init__(self, name, conf, seeding_conf):
        self.name = name
        self.conf = conf
        self.seeding_conf = seeding_conf

        self.coverage = self._coverages()
        self.caches = self._caches()
        self.grids = self._grids(self.caches)
        self.levels = levels_from_options(conf)

    def _coverages(self):
        coverage = None
        if 'coverages' in self.conf:
            try:
                coverages = [self.seeding_conf.coverage(c) for c in self.conf.get('coverages', {})]
            except EmptyCoverageError:
                return False
            if len(coverages) == 1:
                coverage = coverages[0]
            else:
                coverage = MultiCoverage(coverages)
        return coverage

    def _grids(self, caches):
        grids = []

        if 'grids' in self.conf:
            # grids available for all caches
            available_grids = reduce(operator.and_, (set(cache) for cache in caches.values()))
            for grid_name in self.conf['grids']:
                if grid_name not in available_grids:
                    raise SeedConfigurationError('%s not defined for caches' % grid_name)
                grids.append(grid_name)
        else:
            # check that all caches have the same grids configured
            last = []
            for cache_grids in [cache.keys() for cache in caches.values()]:
                if not last:
                    last = cache_grids
                else:
                    if last != cache_grids:
                        raise SeedConfigurationError('caches in same seed task require identical grids')
            grids = list(last or [])
        return grids

    def _caches(self):
        """
        Returns a dictionary with all caches for this seed.

        e.g.: {'seed1': {'grid1': tilemanager1, 'grid2': tilemanager2}}
        """
        caches = {}
        for cache_name in self.conf.get('caches', []):
            caches[cache_name] = self.seeding_conf.cache(cache_name)
        return caches


class SeedConfiguration(ConfigurationBase):
    def __init__(self, name, conf, seeding_conf):
        ConfigurationBase.__init__(self, name, conf, seeding_conf)

        self.refresh_all = False
        self.refresh_timestamp = None
        if 'refresh_before' in self.conf:
            self.refresh_timestamp = before_timestamp_from_options(self.conf['refresh_before'])
        else:
            self.refresh_all = True

    def seed_tasks(self):
        for grid_name in self.grids:
            for cache_name, cache in self.caches.items():
                tile_manager = cache[grid_name]
                grid = self.seeding_conf.grids[grid_name]
                if self.coverage is False:
                    coverage = False
                elif self.coverage:
                    coverage = self.coverage.transform_to(grid.srs)
                else:
                    coverage = BBOXCoverage(grid.bbox, grid.srs)

                try:
                    if coverage is not False:
                        coverage.extent.llbbox
                except TransformationError:
                    raise SeedConfigurationError('%s: coverage transformation error' % self.name)

                if self.levels:
                    levels = self.levels.for_grid(grid)
                else:
                    levels = list(range(0, grid.levels))

                if not tile_manager.cache.supports_timestamp:
                    self.refresh_all = True

                md = dict(name=self.name, cache_name=cache_name, grid_name=grid_name)

                if tile_manager.rescale_tiles:
                    if tile_manager.rescale_tiles > 0:
                        levels = levels[::-1]
                    for level in levels:
                        yield SeedTask(md, tile_manager, [level], self.refresh_timestamp, self.refresh_all, coverage)
                else:
                    yield SeedTask(md, tile_manager, levels, self.refresh_timestamp, self.refresh_all, coverage)


class CleanupConfiguration(ConfigurationBase):
    def __init__(self, name, conf, seeding_conf):
        ConfigurationBase.__init__(self, name, conf, seeding_conf)
        self.init_time = time.time()

        self.remove_all = False
        self.remove_timestamp = self.init_time  # this should not remove
        # fresh seeded tiles, since this should be configured before seeding

        if self.conf.get('remove_all') is True:
            self.remove_all = True
        elif 'remove_before' in self.conf:
            self.remove_timestamp = before_timestamp_from_options(self.conf['remove_before'])

    def cleanup_tasks(self):
        for grid_name in self.grids:
            for cache_name, cache in self.caches.items():
                tile_manager = cache[grid_name]
                grid = self.seeding_conf.grids[grid_name]
                if self.coverage is False:
                    coverage = False
                    complete_extent = False
                elif self.coverage:
                    coverage = self.coverage.transform_to(grid.srs)
                    complete_extent = False
                else:
                    coverage = BBOXCoverage(grid.bbox, grid.srs)
                    complete_extent = True

                try:
                    if coverage is not False:
                        coverage.extent.llbbox
                except TransformationError:
                    raise SeedConfigurationError('%s: coverage transformation error' % self.name)

                if self.levels:
                    levels = self.levels.for_grid(grid)
                else:
                    levels = list(range(0, grid.levels))

                if not tile_manager.cache.supports_timestamp:
                    # for caches without timestamp support (like MBTiles)
                    if self.remove_timestamp is self.init_time:
                        # remove everything
                        self.remove_all = True
                    else:
                        raise SeedConfigurationError(
                            "cleanup does not support remove_before for '%s'"
                            " because cache '%s' does not support timestamps" % (self.name, cache_name))
                md = dict(name=self.name, cache_name=cache_name, grid_name=grid_name)
                yield CleanupTask(md, tile_manager, levels, self.remove_timestamp, remove_all=self.remove_all,
                                  coverage=coverage, complete_extent=complete_extent)


def levels_from_options(conf):
    levels = conf.get('levels')
    if levels:
        if isinstance(levels, list):
            return LevelsList(levels)
        from_level = levels.get('from')
        to_level = levels.get('to')
        return LevelsRange((from_level, to_level))
    resolutions = conf.get('resolutions')
    if resolutions:
        if isinstance(resolutions, list):
            return LevelsResolutionList(resolutions)
        from_res = resolutions.get('from')
        to_res = resolutions.get('to')
        return LevelsResolutionRange((from_res, to_res))
    return None


def before_timestamp_from_options(conf):
    """
    >>> import time
    >>> t = before_timestamp_from_options({'hours': 4})
    >>> time.time() - t - 4 * 60 * 60 < 1
    True
    """
    if 'time' in conf:
        try:
            return timestamp_from_isodate(conf['time'])
        except ValueError:
            raise SeedConfigurationError(
                "can't parse time '%s'. should be ISO time string" % (conf["time"], ))
    if 'mtime' in conf:
        datasource = abspath(conf['mtime'])
        try:
            return os.path.getmtime(datasource)
        except OSError as ex:
            raise SeedConfigurationError(
                "can't parse last modified time from file '%s'." % (datasource, ), ex)
    deltas = {}
    for delta_type in ('weeks', 'days', 'hours', 'minutes', 'seconds'):
        deltas[delta_type] = conf.get(delta_type, 0)
    return timestamp_before(**deltas)


class LevelsList(object):
    def __init__(self, levels=None):
        self.levels = levels

    def for_grid(self, grid):
        uniqe_valid_levels = set(x for x in self.levels if 0 <= x <= (grid.levels-1))
        return sorted(uniqe_valid_levels)


class LevelsRange(object):
    def __init__(self, level_range=None):
        self.level_range = level_range

    def for_grid(self, grid):
        start, stop = self.level_range
        if start is None:
            start = 0
        if stop is None:
            stop = 999

        stop = min(stop, grid.levels-1)

        return list(range(start, stop+1))


class LevelsResolutionRange(object):
    def __init__(self, res_range=None):
        self.res_range = res_range

    def for_grid(self, grid):
        start, stop = self.res_range
        if start is None:
            start = 0
        else:
            start = grid.closest_level(start)

        if stop is None:
            stop = grid.levels-1
        else:
            stop = grid.closest_level(stop)

        return list(range(start, stop+1))


class LevelsResolutionList(object):
    def __init__(self, resolutions=None):
        self.resolutions = resolutions

    def for_grid(self, grid):
        levels = set(grid.closest_level(res) for res in self.resolutions)
        return sorted(levels)



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Service exception handling (WMS exceptions, XML, in_image, etc.).
"""
from html import escape

from mapproxy.response import Response
from mapproxy.template import template_loader
import mapproxy.service
get_template = template_loader(mapproxy.service.__package__, 'templates')


class RequestError(Exception):
    """
    Exception for all request related errors.

    :ivar internal: True if the error was an internal error, ie. the request itself
                    was valid (e.g. the source server is unreachable
    """

    def __init__(self, message, code=None, request=None, internal=False, status=None, locator=None):
        Exception.__init__(self, message)
        self.msg = message
        self.code = code
        self.request = request
        self.internal = internal
        self.status = status
        self.locator = locator

    def render(self):
        """
        Return a response with the rendered exception.
        The rendering is delegated to the ``exception_handler`` that issued
        the ``RequestError``.

        :rtype: `Response`
        """
        if self.request is not None:
            handler = self.request.exception_handler
            resp = handler.render(self)
        elif self.status is not None:
            resp = Response(self.msg, status=self.status)
        else:
            resp = Response('internal error: %s' % self.msg, status=500)
        resp.cache_headers(no_cache=True)
        return resp

    def __str__(self):
        return 'RequestError("%s", code=%r, request=%r)' % (self.msg, self.code,
                                                            self.request)


class ExceptionHandler(object):
    """
    Base class for exception handler.
    """

    def render(self, request_error):
        """
        Return a response with the rendered exception.

        :param request_error: the exception to render
        :type request_error: `RequestError`
        :rtype: `Response`
        """
        raise NotImplementedError()


def _not_implemented(*args, **kw):
    raise NotImplementedError()


class XMLExceptionHandler(ExceptionHandler):
    """
    Mixin class for tempita-based template renderer.
    """
    template_file = None
    """The filename of the tempita xml template"""

    content_type = None
    """
    The mime type of the exception response (use this or mimetype).
    The content_type is sent as defined here.
    """

    status_code = 500
    """
    The HTTP status code.
    """

    status_codes = {}
    """
    Mapping of exceptionCodes to status_codes. If not defined
    status_code is used.
    """

    mimetype = None
    """
    The mime type of the exception response. (use this or content_type).
    A character encoding might be added to the mimetype (like text/xml;charset=UTF-8)
    """

    template_func = _not_implemented
    """
    Function that returns the named template.
    """

    def render(self, request_error):
        """
        Render the template of this exception handler. Passes the
        ``request_error.msg`` and ``request_error.code`` to the template.

        :type request_error: `RequestError`
        """
        if request_error.status is not None:
            status_code = request_error.status
        else:
            status_code = self.status_codes.get(request_error.code, self.status_code)

        # escape &<> in error message (e.g. URL params)
        msg = escape(request_error.msg)
        result = self.template.substitute(exception=msg,
                                          code=request_error.code)
        return Response(result, mimetype=self.mimetype, content_type=self.content_type,
                        status=status_code)

    @property
    def template(self):
        """
        The template for this ExceptionHandler.
        """
        return self.template_func(self.template_file)


class OWSExceptionHandler(XMLExceptionHandler):
    """
    Exception handler for generic OWS ServiceExceptionReports
    """
    template_file = 'ows_exception.xml'
    template_func = get_template
    mimetype = 'text/xml'

    def render(self, request_error):
        """
        Render the template of this exception handler. Passes the
        ``request_error.msg``, ``request_error.locator`` and ``request_error.code`` to the template.

        :type request_error: `RequestError`
        """
        if request_error.status is not None:
            status_code = request_error.status
        else:
            status_code = self.status_codes.get(request_error.code, self.status_code)

        # escape &<> in error message (e.g. URL params)
        msg = escape(request_error.msg)
        result = self.template.substitute(exception=msg,
                                          code=request_error.code, locator=request_error.locator)
        return Response(result, mimetype=self.mimetype, content_type=self.content_type,
                        status=status_code)


class PlainExceptionHandler(ExceptionHandler):
    mimetype = 'text/plain'
    status_code = 404

    def render(self, request_error):
        if request_error.internal:
            self.status_code = 500
        return Response(request_error.msg, status=self.status_code,
                        mimetype=self.mimetype)



from ..util import resolve_ns


def test_resolve_ns():
    assert resolve_ns("/bar/bar", {}, None) == "/bar/bar"

    assert (
        resolve_ns("/bar/bar", {}, "http://foo") == "/{http://foo}bar/{http://foo}bar"
    )

    assert (
        resolve_ns(
            "/bar/xlink:bar", {"xlink": "http://www.w3.org/1999/xlink"}, "http://foo"
        )
        == "/{http://foo}bar/{http://www.w3.org/1999/xlink}bar"
    )

    assert (
        resolve_ns(
            "bar/xlink:bar", {"xlink": "http://www.w3.org/1999/xlink"}, "http://foo"
        )
        == "{http://foo}bar/{http://www.w3.org/1999/xlink}bar"
    )



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO

from mapproxy.client.arcgis import ArcGISInfoClient
from mapproxy.layer import InfoQuery
from mapproxy.request.arcgis import ArcGISIdentifyRequest
from mapproxy.srs import SRS, SupportedSRS
from mapproxy.test.http import assert_query_eq


TESTSERVER_ADDRESS = ('127.0.0.1', 56413)
TESTSERVER_URL = 'http://%s:%s' % TESTSERVER_ADDRESS


class MockHTTPClient(object):
    def __init__(self):
        self.requested = []

    def open(self, url, data=None):
        self.requested.append(url)
        result = BytesIO(b'{}')
        result.seek(0)
        result.headers = {}
        return result


class TestArcGISInfoClient(object):
    def test_fi_request(self):
        req = ArcGISIdentifyRequest(url=TESTSERVER_URL + '/MapServer/export?map=foo', param={'layers': 'foo'})
        http = MockHTTPClient()
        wms = ArcGISInfoClient(req, http_client=http, supported_srs=SupportedSRS([SRS(4326)]))
        fi_req = InfoQuery((8, 50, 9, 51), (512, 512),
                           SRS(4326), (128, 64), 'text/plain')

        wms.get_info(fi_req)

        assert_query_eq(http.requested[0],
                        TESTSERVER_URL+'/MapServer/identify?map=foo'
                        '&imageDisplay=512,512,96&sr=4326&f=json'
                        '&layers=foo&tolerance=5&returnGeometry=false'
                        '&geometryType=esriGeometryPoint&geometry=8.250000,50.875000'
                        '&mapExtent=8,50,9,51',
                        fuzzy_number_compare=True)

    def test_transform_fi_request_supported_srs(self):
        req = ArcGISIdentifyRequest(url=TESTSERVER_URL + '/MapServer/export?map=foo', param={'layers': 'foo'})
        http = MockHTTPClient()
        wms = ArcGISInfoClient(req, http_client=http, supported_srs=SupportedSRS([SRS(25832)]))
        fi_req = InfoQuery((8, 50, 9, 51), (512, 512),
                           SRS(4326), (128, 64), 'text/plain')

        wms.get_info(fi_req)

        assert_query_eq(http.requested[0],
                        TESTSERVER_URL+'/MapServer/identify?map=foo'
                        '&imageDisplay=512,797,96&sr=25832&f=json'
                        '&layers=foo&tolerance=5&returnGeometry=false'
                        '&geometryType=esriGeometryPoint&geometry=447229.979084,5636149.370634'
                        '&mapExtent=428333.552496,5538630.70275,500000.0,5650300.78652',
                        fuzzy_number_compare=True)
